#include <stdio.h>
#include "Database.h"
#include "GetOpt.h"
#include "GenerateEnv.h"

char g_pszDBSHost[100];
char g_pszDBSPort[100];

//////////////////////////////////////////////
// These variables and function are provided
// so I can use NT_Tcp_sockets.cpp
bool g_bVerbose = false;
int g_nIproc = 0;
int g_nNproc = 1;
void MakeErrMsg(int error, char *pFormat, ...)
{
	char chMsg[1024];
	va_list pArg;
	
	va_start(pArg, pFormat);
	vsprintf(chMsg, pFormat, pArg);
	va_end(pArg);

	printf("Error %d:\n   %s\n", error, chMsg);
    
	WSACleanup();
	ExitProcess(1);
}
//////////////////////////////////////////////

void PrintOptions()
{
	printf("Usage:\n");
	printf(" mpichdbs\n");
	printf(" mpichdbs -port port\n");
	printf(" mpichdbs -feed file host port\n");
	printf(" mpichdbs -connect host port id\n");
	//printf(" mpichdbs -print host port\n");
}

void PrintHelp()
{
	printf("commands:\n");
	printf(" port\n");
	printf(" host\n");
	printf(" feed filename\n");
	printf(" print\n");
	printf(" set id\n");
	printf(" put key=value\n");
	printf(" putc key=value\n");
	printf(" get key\n");
	printf(" delete node\n");
	printf(" delete\n");
	printf(" help\n");
	printf(" quit\n");
	printf("\n");
}

void PrintRemoteHelp()
{
	printf("commands:\n");
	printf(" feed filename\n");
	printf(" print\n");
	printf(" set id\n");
	printf(" put key=value\n");
	printf(" putc key=value\n");
	printf(" get key\n");
	printf(" delete node\n");
	printf(" delete\n");
	printf(" help\n");
	printf(" quit\n");
	printf("\n");
}

void Feed(DatabaseServer &dbs, char *pszFilename)
{
	int i=0;
	char pszKey[100];
	char **ppEnvStrings=NULL, **ppIter;

	dbs.GetHost(g_pszDBSHost, 100);
	sprintf(g_pszDBSPort, "%d", dbs.GetPort());
	ppEnvStrings = GenerateEnvironmentStrings(pszFilename);
	if (ppEnvStrings == NULL)
		return;

	// Delete any old lingering data before adding new stuff to the database
	dbs.Delete(g_pszJobID);

	ppIter = ppEnvStrings;
	while (*ppIter != NULL)
	{
		if (g_bStaticEnvironment)
			strcpy(pszKey, "env");
		else
			sprintf(pszKey, "env%d", i);
		char *pKey = new char[strlen(pszKey)+1];
		strcpy(pKey, pszKey);

		if (dbs.Put(g_pszJobID, pKey, *ppIter, strlen(*ppIter)+1, !g_bStaticEnvironment) == MPI_DBS_FAIL)
		{
			printf("DatabaseServer.Put failed\n");
			return;
		}

		ppIter++;
		i++;
	}
	if (ppEnvStrings != NULL)
		delete ppEnvStrings;
}

void Feed(Database &db, char *pszFilename)
{
	int i=0;
	char pszKey[100];
	char **ppEnvStrings=NULL, **ppIter;

	ppEnvStrings = GenerateEnvironmentStrings(pszFilename);
	if (ppEnvStrings == NULL)
		return;

	db.SetID(g_pszJobID);

	// Delete any old lingering data before adding new stuff to the database
	db.Delete();

	ppIter = ppEnvStrings;
	while (*ppIter != NULL)
	{
		if (g_bStaticEnvironment)
			strcpy(pszKey, "env");
		else
			sprintf(pszKey, "env%d", i);
		if (db.Put(pszKey, *ppIter, strlen(*ppIter)+1, !g_bStaticEnvironment) == MPI_DBS_FAIL)
		{
			printf("Database.Put failed\n");
			return;
		}
		delete *ppIter;

		ppIter++;
		i++;
	}
	if (ppEnvStrings != NULL)
		delete ppEnvStrings;
}

void main(int argc, char *argv[])
{
	DatabaseServer dbs;
	Database db;
	int nPort = 0;
	char pszFilename[MAX_PATH];
	char pszHost[100];
	char cmd[1024] = "init";
	bool bSetCalled = false;
	char pszID[1024];
	char pszKey[1024], pszValue[1024];

	if (GetOpt(argc, argv, "-help") || GetOpt(argc, argv, "-?"))
	{
		PrintOptions();
		return;
	}

	// Start the Winsock dll
	WSADATA wsaData;
	int err = WSAStartup( MAKEWORD( 2, 0 ), &wsaData );
	if (err != 0)
		printf("Winsock2 dll not initialized\n");

	GetOpt(argc, argv, "-port", &nPort);

	if (GetOpt(argc, argv, "-feed", pszFilename))
	{
		if (argc < 3)
		{
			PrintOptions();
			return;
		}
		strcpy(pszHost, argv[1]);
		nPort = atoi(argv[2]);

		strcpy(g_pszDBSHost, pszHost);
		strcpy(g_pszDBSPort, argv[2]);
		SetEnvironmentVariable("MPICH_DBS_HOST", pszHost);
		SetEnvironmentVariable("MPICH_DBS_PORT", argv[2]);
		if (db.Init())
			Feed(db, pszFilename);
		else
			printf("Database initialization failed\n");

		return;
	}

	/*
	if (GetOpt(argc, argv, "-print", pszHost))
	{
		if (argc < 2)
		{
			PrintOptions();
			return;
		}
		nPort = atoi(argv[1]);

		// Add print code here

		return;
	}
	//*/

	if (GetOpt(argc, argv, "-connect", pszHost))
	{
		if (argc < 3)
		{
			PrintOptions();
			return;
		}
		strcpy(pszID, argv[2]);
		strcpy(g_pszDBSHost, pszHost);
		strcpy(g_pszDBSPort, argv[1]);
		SetEnvironmentVariable("MPICH_DBS_HOST", pszHost);
		SetEnvironmentVariable("MPICH_DBS_PORT", argv[1]);
		db.SetID(argv[2]);
		if (!db.Init())
		{
			printf("Database initialization failed\n");
			return;
		}
		printf("MPICH Database connection to '%s' on port %d with id=\"%s\"\n", pszHost, atoi(argv[1]), pszID);
		PrintRemoteHelp();
		printf(":");
		gets(cmd);
		while ((stricmp(cmd, "quit") != 0) && (stricmp(cmd, "exit") != 0))
		{
			if (strnicmp(cmd, "set ", 4) == 0)
			{
				strcpy(pszID, &cmd[4]);
				db.SetID(pszID);
			}
			else
			if (strnicmp(cmd, "put ", 4) == 0)
			{
				char *token;
				token = strtok(&cmd[4], "=");
				if (token != NULL)
					strcpy(pszKey, token);
				token = strtok(NULL, "\n");
				if (token != NULL)
					strcpy(pszValue, token);
				if (db.Put(pszKey, pszValue, strlen(pszValue)+1) == MPI_DBS_FAIL)
					printf("Database.Put failed\n");
			}
			else
			if (strnicmp(cmd, "putc ", 5) == 0)
			{
				char *token;
				token = strtok(&cmd[5], "=");
				if (token != NULL)
					strcpy(pszKey, token);
				token = strtok(NULL, "\n");
				if (token != NULL)
					strcpy(pszValue, token);
				if (db.Put(pszKey, pszValue, strlen(pszValue)+1, false) == MPI_DBS_FAIL)
					printf("Database.Put failed\n");
			}
			else
			if (strnicmp(cmd, "get ", 4) == 0)
			{
				int length = 1024;
				if (db.Get(&cmd[4], pszValue, &length) == MPI_DBS_SUCCESS)
				{
					fwrite(pszValue, length, 1, stdout);
					printf("\n");
				}
				else
					printf("Database.Get failed\n");
			}
			else
			if (strnicmp(cmd, "delete ", 7) == 0)
			{
				char pszTemp[100];
				int length = 100;
				if (db.GetID(pszTemp, &length))
				{
					db.SetID(&cmd[7]);
					if (db.Delete() == MPI_DBS_SUCCESS)
						printf("node '%s' deleted from the database\n", &cmd[7]);
					else
						printf("failed to delete node '%s' from the database\n", &cmd[7]);
					db.SetID(pszTemp);
				}
			}
			else
			if (stricmp(cmd, "delete") == 0)
			{
				if (db.Delete() == MPI_DBS_SUCCESS)
					printf("node '%s' deleted from the database\n", pszID);
				else
					printf("failed to delete node '%s' from the database\n", pszID);
			}
			else
			if (strnicmp(cmd, "feed ", 5) == 0)
				Feed(db, &cmd[5]);
			else
			if (stricmp(cmd, "print") == 0)
			{
				char pBuffer[4096];
				int length = 4096;
				if (db.Print(pBuffer, &length) == MPI_DBS_SUCCESS)
					fwrite(pBuffer, length-1, 1, stdout);
				else
					printf("Database.Print failed\n");
			}
			else
			if (stricmp(cmd, "help") == 0)
				PrintRemoteHelp();
			else
			if (stricmp(cmd, "id") == 0)
			{
				printf("id = %s\n", pszID);
			}
			else
			{
				if (strlen(cmd))
					printf("Unknown command\n");
			}
			printf(":");
			gets(cmd);
		}
		return;
	}

	// If we make it here then we are set to run the database server and wait for commands
	if (!dbs.SetPort(nPort))
	{
		printf("Unable to set the port\n");
		return;
	}
	if (!dbs.Start())
	{
		printf("Unable to start the database server thread\n");
		return;
	}

	dbs.GetHost(pszHost, 100);
	printf("MPICH Database started on '%s' port: %d\n", pszHost, dbs.GetPort());
	PrintHelp();
	printf(":");
	gets(cmd);
	while ((stricmp(cmd, "quit") != 0) && (stricmp(cmd, "exit") != 0))
	{
		if (stricmp(cmd, "print") == 0)
		{
			dbs.PrintState();
		}
		else
		if (strnicmp(cmd, "set ", 4) == 0)
		{
			strcpy(pszID, &cmd[4]);
			bSetCalled = true;
		}
		else
		if (strnicmp(cmd, "put ", 4) == 0)
		{
			if (bSetCalled)
			{
				char *token;
				token = strtok(&cmd[4], "=");
				if (token != NULL)
					strcpy(pszKey, token);
				token = strtok(NULL, "\n");
				if (token != NULL)
					strcpy(pszValue, token);
				char *pKey = new char[strlen(pszKey)+1];
				strcpy(pKey, pszKey);
				char *pValue = new char[strlen(pszValue)+1];
				strcpy(pValue, pszValue);
				if (dbs.Put(pszID, pKey, pValue, strlen(pszValue)+1) == MPI_DBS_FAIL)
					printf("DatabaseServer.Put failed\n");
			}
			else
				printf("'set id' must be called before put\n");
		}
		else
		if (strnicmp(cmd, "putc ", 5) == 0)
		{
			if (bSetCalled)
			{
				char *token;
				token = strtok(&cmd[5], "=");
				if (token != NULL)
					strcpy(pszKey, token);
				token = strtok(NULL, "\n");
				if (token != NULL)
					strcpy(pszValue, token);
				char *pKey = new char[strlen(pszKey)+1];
				strcpy(pKey, pszKey);
				char *pValue = new char[strlen(pszValue)+1];
				strcpy(pValue, pszValue);
				if (dbs.Put(pszID, pKey, pValue, strlen(pValue)+1, false) == MPI_DBS_FAIL)
					printf("DatabaseServer.Put failed\n");
			}
			else
				printf("'set id' must be called before put\n");
		}
		else
		if (strnicmp(cmd, "get ", 4) == 0)
		{
			if (bSetCalled)
			{
				int length = 1024;
				void *pValue;
				if (dbs.Get(pszID, &cmd[4], pValue, &length) == MPI_DBS_SUCCESS)
				{
					fwrite(pValue, length, 1, stdout);
					delete pValue;
					printf("\n");
				}
				else
					printf("DatabaseServer.Get failed\n");
			}
			else
				printf("'set id' must be called before get\n");
		}
		else
		if (stricmp(cmd, "id") == 0)
		{
			if (bSetCalled)
				printf("id = %s\n", pszID);
			else
				printf("call 'set id' before calling 'id'\n");
		}
		else
		if (stricmp(cmd, "delete") == 0)
		{
			if (bSetCalled)
			{
				if (dbs.Delete(pszID) == MPI_DBS_SUCCESS)
					printf("node '%s' deleted from the database\n", pszID);
				else
					printf("failed to delete node '%s' from the database\n", pszID);
			}
			else
				printf("no node deleted\n'set id' must be called before delete or call 'delete node' directly\n");
		}
		else
		if (strnicmp(cmd, "delete ", 7) == 0)
		{
			if (dbs.Delete(&cmd[7]) == MPI_DBS_SUCCESS)
				printf("node '%s' deleted from the database\n", &cmd[7]);
			else
				printf("failed to delete node '%s' from the database\n", &cmd[7]);
		}
		else
		if (stricmp(cmd, "port") == 0)
		{
			printf("Port: %d\n", dbs.GetPort());
		}
		else
		if (stricmp(cmd, "host") == 0)
		{
			dbs.GetHost(pszHost, 100);
			printf("Host: %s\n", pszHost);
		}
		else
		if (strnicmp(cmd, "feed ", 5) == 0)
			Feed(dbs, &cmd[5]);
		else
		if (stricmp(cmd, "help") == 0)
			PrintHelp();
		else
		{
			if (strlen(cmd))
				printf("Unknown command\n");
		}
		printf(":");
		gets(cmd);
	}
	dbs.Stop();
	WSACleanup();
}
